using JET, InteractiveUtils

const CC = JET.CC

import .CC:
    Bottom, widenconst, ⊑

import JET:
    AbstractAnalyzer, AbstractGlobal, InferenceErrorReport, JETAnalyzer, ToplevelConfig,
    ToplevelErrorReport, gen_virtual_module, get_reports, get_result, print_reports,
    virtual_process, virtualize_module_context

import Base.Meta:
    isexpr

get_cache(analyzer::AbstractAnalyzer) = JET.get_inf_cache(analyzer)

function subtypes_recursive!(t, ts)
    push!(ts, t)
    if isabstracttype(t)
        for ct in subtypes(t)
            subtypes_recursive!(ct, ts)
        end
    end
    return ts
end

let ts = Type[]
    subtypes_recursive!(ToplevelErrorReport, ts)
    subtypes_recursive!(InferenceErrorReport, ts)
    for t in ts
        canonicalname = Symbol(parentmodule(t), '.', nameof(t))
        canonicalpath = Symbol.(split(string(canonicalname), '.'))

        modpath = Expr(:., canonicalpath[1:end-1]...)
        symname = Expr(:., last(canonicalpath))
        ex = Expr(:import, Expr(:(:), modpath, symname))
        Core.eval(@__MODULE__, ex)
    end
end

iskwarg(@nospecialize(x)) = isexpr(x, :(=))

"""
    @fixturedef

Creates a virtual module and defines fixtures from toplevel expression `ex`
"""
macro fixturedef(ex)
    @assert isexpr(ex, :block)
    return quote let
        vmod = $gen_virtual_module()
        for x in $(ex.args)
            Core.eval(vmod, x)
        end
        vmod # return virtual module
    end end
end

"""
    @analyze_toplevel [jetconfigs...] ex -> JET.VirtualProcessResult

Enters JET analysis from toplevel expression `ex`, and returns the analysis result.
"""
macro analyze_toplevel(xs...)
    jetconfigs = filter(iskwarg, xs)
    xs′ = filter(!iskwarg, xs)
    @assert length(xs′) == 1
    ex = first(xs′)
    return _analyze_toplevel(ex, __source__, jetconfigs)
end

"""
    @analyze_toplevel2 [jetconfigs...] ex -> (Module, JET.VirtualProcessResult)

Works similarly to `@analyze_toplevel`, but also creates a virtual module beforehand and
returns that as well after the whole analysis for the later inspection.
"""
macro analyze_toplevel2(xs...)
    jetconfigs = filter(iskwarg, xs)
    xs′ = filter(!iskwarg, xs)
    @assert length(xs′) == 1
    ex = first(xs′)

    vmod = gensym(:vmod)
    jetconfigs = (:(context = $vmod), jetconfigs...,)
    jetconfigs = (:(virtualize = false), jetconfigs...,)
    ex2 = _analyze_toplevel(ex, __source__, jetconfigs)
    return :(let
        $(esc(vmod)) = $gen_virtual_module()
        ret2 = $ex2
        $(esc(vmod)), ret2
    end)
end

function _analyze_toplevel(ex, lnn, jetconfigs)
    toplevelex = (isexpr(ex, :block) ?
                  Expr(:toplevel, lnn, ex.args...) : # flatten here
                  Expr(:toplevel, lnn, ex)
                  ) |> QuoteNode
    return :(let
        analyzer = $(GlobalRef(JET, :JETAnalyzer))(; $(map(esc, jetconfigs)...))
        config = ToplevelConfig(; $(map(esc, jetconfigs)...))
        res = $virtual_process($toplevelex,
                               $(string(lnn.file)),
                               analyzer,
                               config,
                               )
        JET.JETToplevelResult(analyzer, res, "analyze_toplevel"; $(map(esc, jetconfigs)...))
    end)
end

# `report_file` with silent top-level logger
report_file2(args...; kwargs...) = report_file(args...; toplevel_logger = nothing, kwargs...)

is_concrete(mod::Module, sym::Symbol) = isdefined(mod, sym) && !isa(JET.getglobal(mod, sym), AbstractGlobal)
isa_concrete(mod::Module, sym::Symbol, @nospecialize(typ)) = is_concrete(mod, sym) && isa(JET.getglobal(mod, sym), typ)

is_abstract(mod::Module, sym::Symbol) = isdefined(mod, sym) && isa(JET.getglobal(mod, sym), AbstractGlobal)
isa_abstract(mod::Module, sym::Symbol, @nospecialize(typ)) = is_abstract(mod, sym) && (JET.getglobal(mod, sym)::AbstractGlobal).t ⊑ typ

# JET will try to concretize global variable when its type is a constant at analysis time,
# but the starategy is a bit complicated right now and may change in the future
# these utilities allow robust testing to check if a object is successfully analyzed by JET whichever it's concretized or abstracted
is_analyzed(mod::Module, sym::Symbol) = isdefined(mod, sym) # essentially, `is_concrete(mod, sym) || is_abstract(mod, sym)`
isa_analyzed(mod::Module, sym::Symbol, @nospecialize(typ)) = isa_abstract(mod, sym, typ) || isa_concrete(mod, sym, typ)

function is_global_undef_var(@nospecialize(r::InferenceErrorReport), mod::Module, name::Symbol)
    r isa UndefVarErrorReport || return false
    var = r.var
    return var isa GlobalRef && var.mod === mod && var.name === name
end
function is_global_undef_var(@nospecialize(r::InferenceErrorReport), name::Symbol)
    r isa UndefVarErrorReport || return false
    var = r.var
    return var isa GlobalRef && var.name === name
end

function is_local_undef_var(@nospecialize(r::InferenceErrorReport), name::Symbol)
    r isa UndefVarErrorReport || return false
    var = r.var
    return var === name
end
